// LCD_KS0108.c

//----------------------------------------------------------------------------
// Universal KS0108 driver library
// (c) Radosaw Kwiecie, radek@dxp.pl
//----------------------------------------------------------------------------


//============================================================================
//				HEADERS
//============================================================================

#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "LCD_KS0108.h"
#include "EvoBox.h"
#include "Fonts.h"


//============================================================================
//				DEFINITIONS
//============================================================================

#define KS0108_DATA_PORT	PORTA
#define KS0108_DATA_DIR		DDRA
#define KS0108_DATA_PIN		PINA
#define KS0108_CTRL_PORT	PORTB
#define KS0108_CTRL_DIR		DDRB
#define KS0108_DI			0x10
#define KS0108_RW			0x08
#define KS0108_EN			0x04
#define KS0108_CS1			0x02
#define KS0108_CS2			0x01
#define DISPLAY_STATUS_BUSY 0x80


//============================================================================
//				GLOBAL VARIABLES
//============================================================================

unsigned char screen_x = 0, screen_y = 0;
volatile unsigned char LCDInvert = 0;
volatile unsigned char Font_Bold = 0;
volatile unsigned char FontS_Offset = 0;
         char StringBuffer[33],NumberBuffer[33];

#ifdef CFG_SCROLL_SCREEN_ENABLE
	unsigned char ScreenBuffer[SCAN_SCREEN_WIDTH_X_MAX][SCAN_SCREEN_HEIGHT_Y_MAX];
#endif


//============================================================================
//				FUNCTION PROTOTYPES
//============================================================================

// Basic LCD Functions
void GLCD_EnableController(unsigned char controller);
void GLCD_DisableController(unsigned char controller);
unsigned char GLCD_ReadStatus(unsigned char controller);
unsigned char GLCD_ReadData(void);
void GLCD_WriteCommand(unsigned char commandToWrite, unsigned char controller);
void GLCD_WriteData(unsigned char dataToWrite);

// User LCD Functions
void GLCD_Initalize(void);
void GLCD_GoTo(unsigned char x, unsigned char y);
void GLCD_FillScreen(unsigned char pFill);
void GLCD_SetPixel(unsigned char x, unsigned char y, unsigned char color);

// Small 5x4 Font Functions
void GLCD_WriteCharS(unsigned char pCharToWrite);
void GLCD_WriteStringS(char* pStringToWrite);
void GLCD_WriteStringS_P(const char* pStringToWrite);
void GLCD_WriteNumberS(signed short pNumber);
void GLCD_WriteNumberS_Long(signed long pNumber);

// Medium 7 Font Functions
void GLCD_WriteCharM(unsigned char pCharToWrite);
void GLCD_WriteStringM(char* pStringToWrite);
void GLCD_WriteStringM_P(const char* pStringToWrite);
void GLCD_WriteNumberM(signed short pNumber);

// Font "Zero Twos" for Mitsubishi
void GLCD_WriteDigitZero(unsigned char pDigit,unsigned char pFont);
void GLCD_WriteNumberZero(unsigned short pNumber,unsigned char pKomma,unsigned char pFont);
void GLCD_WriteNumberZero_Format(unsigned char pXCursor,unsigned char pYCursor,unsigned char* pNumberAddr,unsigned char pBit,unsigned char pDigits,unsigned char pKomma,unsigned char pFont,unsigned char pDirection);
void GLCD_WriteNumberZero_SFormat(unsigned char pXCursor,unsigned char pYCursor,signed char* pNumberAddr,unsigned char pBit,unsigned char pDigits,unsigned char pKomma,unsigned char pFont,unsigned char pDirection);

// LCD Images Functions
void GLCD_DrawImage(const unsigned char* pImage, unsigned char pPosX, unsigned char pPosY, unsigned char pWidthX, unsigned char pHeightY);
void GLCD_DrawAni(unsigned short* pAni, unsigned char pFrames, unsigned char pPosX, unsigned char pPosY, unsigned char pWidthX, unsigned char pHeightY, unsigned short pWait);
void GLCD_DrawSimpleLine(unsigned char pPosX1,unsigned char pPosY1,unsigned char pPosX2,unsigned char pPosY2, unsigned char pFillByte);
void GLCD_DrawLine(unsigned char x1,unsigned char y1,unsigned char x2,unsigned char y2,unsigned char color,unsigned char pBold);

// Scroll + Scan Functions
#ifdef CFG_SCROLL_SCREEN_ENABLE
	void GLCD_ScrollScreen(unsigned char pDirection,unsigned char pCursorX,unsigned char pPageY,unsigned char pWidthX,unsigned char pHeightY,unsigned char pLines,unsigned short pWait);
	void GLCD_ScanScreen(unsigned char pCursorX,unsigned char pPageY,unsigned char pWidthX,unsigned char pHeightY);
#endif


//============================================================================
//				BASIC LCD-FUNCTIONS
//============================================================================

// Enable Controller (0-1)
void GLCD_EnableController(unsigned char controller)
{
	if(controller == 0) KS0108_CTRL_PORT &= ~KS0108_CS1;
	else KS0108_CTRL_PORT &= ~KS0108_CS2;
}

// Disable Controller (0-1)
void GLCD_DisableController(unsigned char controller)
{
	if(controller == 0) KS0108_CTRL_PORT |= KS0108_CS1;
	else KS0108_CTRL_PORT |= KS0108_CS2;
}

// Read Status from specified controller (0-1)
unsigned char GLCD_ReadStatus(unsigned char controller)
{
	unsigned char status;
	
	KS0108_DATA_DIR = 0x00;
	KS0108_CTRL_PORT |= KS0108_RW;
	KS0108_CTRL_PORT &= ~KS0108_DI;
	
	GLCD_EnableController(controller);
	KS0108_CTRL_PORT |= KS0108_EN;
	status = KS0108_DATA_PIN;
	KS0108_CTRL_PORT &= ~KS0108_EN;
	GLCD_DisableController(controller);

	return status;
}

// Write command to specified controller
void GLCD_WriteCommand(unsigned char commandToWrite, unsigned char controller)
{
	unsigned char i;

	#ifndef _IFDEF_FAST_LCD
		while(GLCD_ReadStatus(controller) & DISPLAY_STATUS_BUSY);
	#endif
	
	KS0108_DATA_DIR = 0xFF;
	KS0108_CTRL_PORT &= ~(KS0108_RW | KS0108_DI);
	
	GLCD_EnableController(controller);
	KS0108_DATA_PORT = commandToWrite;
	KS0108_CTRL_PORT |= KS0108_EN;
	KS0108_CTRL_PORT &= ~KS0108_EN;
	GLCD_DisableController(controller);
	
	GLCD_Delay(30);	// @ 20MHz
}

// Read data from current position
// use it double!
unsigned char GLCD_ReadData(void)
{
	unsigned char data;

	#ifndef _IFDEF_FAST_LCD
		while(GLCD_ReadStatus(screen_x / 64) & DISPLAY_STATUS_BUSY);
	#endif
	
	KS0108_DATA_DIR = 0x00;
	KS0108_CTRL_PORT |= (KS0108_RW | KS0108_DI);
	
	GLCD_EnableController(screen_x / 64);
	KS0108_CTRL_PORT |= KS0108_EN;
	
	if(LCDInvert == 0) data = KS0108_DATA_PIN;
	else data = 0xFF - KS0108_DATA_PIN;

	KS0108_CTRL_PORT &= ~KS0108_EN;
	GLCD_DisableController(screen_x / 64);
		
	screen_x++;
	return data;
}

// Write data to current position
void GLCD_WriteData(unsigned char dataToWrite)
{
	unsigned char i;

	#ifndef _IFDEF_FAST_LCD
		while(GLCD_ReadStatus(screen_x / 64)&DISPLAY_STATUS_BUSY);
	#endif
	
	KS0108_DATA_DIR = 0xFF;
	KS0108_CTRL_PORT &= ~KS0108_RW;
	KS0108_CTRL_PORT |= KS0108_DI;
	
	if(LCDInvert == 0) KS0108_DATA_PORT = dataToWrite;
	else KS0108_DATA_PORT = 0xFF - dataToWrite;

	GLCD_EnableController(screen_x / 64);
	KS0108_CTRL_PORT |= KS0108_EN;
	KS0108_CTRL_PORT &= ~KS0108_EN;
	GLCD_DisableController(screen_x / 64);
	
	GLCD_Delay(30);	// @ 20MHz
	
	screen_x++;
}


//============================================================================
//				USER LCD-FUNCTIONS
//============================================================================

// Initialize
void GLCD_Initalize(void)
{	
	KS0108_CTRL_PORT |= (KS0108_CS1 | KS0108_CS2);
	KS0108_CTRL_DIR |= (KS0108_CS1 | KS0108_CS2 | KS0108_DI | KS0108_RW | KS0108_EN);
	
	GLCD_WriteCommand((DISPLAY_ON_CMD | ON), 0);
	GLCD_WriteCommand((DISPLAY_ON_CMD | ON), 1);
}

// Goto X-Y-Posiion X=[0;128] Y=[0;63]
void GLCD_GoTo(unsigned char x, unsigned char y)
{
	screen_x = x;
	screen_y = y;
	
	GLCD_WriteCommand(DISPLAY_SET_Y | 0,0);
	GLCD_WriteCommand(DISPLAY_SET_X | y,0);
	GLCD_WriteCommand(DISPLAY_START_LINE | 0,0);

	GLCD_WriteCommand(DISPLAY_SET_Y | 0,1);
	GLCD_WriteCommand(DISPLAY_SET_X | y,1);
	GLCD_WriteCommand(DISPLAY_START_LINE | 0,1);

	GLCD_WriteCommand(DISPLAY_SET_Y | (x%64), (x/64));
	GLCD_WriteCommand(DISPLAY_SET_X | y, (x/64));
}

// Clear LCD
void GLCD_FillScreen(unsigned char pFill)
{
	unsigned char i, j;
	
	for(j=0; j<8; j++)
	{
		GLCD_GoTo(0,j);
		for(i=0; i<KS0108_SCREEN_WIDTH; i++) GLCD_WriteData(pFill);
	}
}

// Set Pixel
void GLCD_SetPixel(unsigned char pX, unsigned char pY, unsigned char pColor)
{
	unsigned char tmp;
	GLCD_GoTo(pX, (pY/8));
	GLCD_ReadData();
	
	GLCD_GoTo(pX, (pY/8));
	tmp = GLCD_ReadData();
	
	GLCD_GoTo(pX, (pY/8));
	if(pColor) tmp |= (1 << (pY%8));
	else tmp &= ~(1 << (pY%8));

	GLCD_WriteData(tmp);
}


//============================================================================
//				SMALL 5x4-FONTS FUNCTIONS
//============================================================================

// Write Character with SMALL Font size 5x4
void GLCD_WriteCharS(unsigned char pCharToWrite)
{
	unsigned short FontIndex;
	unsigned char *CharacterROMAdress;

	     if(pCharToWrite >= 'a' && pCharToWrite <= 'z') pCharToWrite -= 32;		// Kleinbuchstaben in Grobuchstaben wandeln
	else if(pCharToWrite >= '{') pCharToWrite -= 26;							// weil Kleinbuchstaben fehlen -> Offset
	
	FontIndex = (pCharToWrite - 32) * 3;
	
	CharacterROMAdress = (unsigned char*)(Font5x4 + FontIndex);

	GLCD_WriteData(0x00);
	GLCD_WriteData( (pgm_read_byte(CharacterROMAdress  )<<FontS_Offset) );
	GLCD_WriteData( (pgm_read_byte(CharacterROMAdress+1)<<FontS_Offset) );
	GLCD_WriteData( (pgm_read_byte(CharacterROMAdress+2)<<FontS_Offset) );
}

// Write String with SMALL Font size 5x4
void GLCD_WriteStringS(char* pStringToWrite)
{
	while(*pStringToWrite)
	{
		GLCD_WriteCharS((char)(*pStringToWrite));
		pStringToWrite++;
	}	
}	

// Write String from FLASH with SMALL Font size 5x4
void GLCD_WriteStringS_P(const char* pStringToWrite)
{
	register char Character;

	while( (Character = pgm_read_byte(pStringToWrite)) )
	{
		GLCD_WriteCharS(Character);
		pStringToWrite++;
	}
}	

// Write Integer with SMALL Font size 5x4
void GLCD_WriteNumberS(signed short pNumber)
{
	itoa(pNumber,NumberBuffer,10);
	GLCD_WriteStringS(NumberBuffer);
}

// Write Integer with SMALL Font size 5x4 DATATYPE LONG
void GLCD_WriteNumberS_Long(signed long pNumber)
{
	ltoa(pNumber,NumberBuffer,10);
	GLCD_WriteStringS(NumberBuffer);
}


//============================================================================
//				MEDIUM 7-FONTS FUNCTIONS
//============================================================================

// Write Character with MEDIUM Font size 7
void GLCD_WriteCharM(unsigned char pCharToWrite)
{
	unsigned short FontIndex;
	unsigned char *CharacterROMAdress;
	unsigned char i,FontWidthIndex;
	
	// Decide between Normal and Bold
	if(Font_Bold == 0)
	{
		     if(pCharToWrite == 0x0A) pCharToWrite += 120;
		else if(pCharToWrite == 0x0D) pCharToWrite += 116;

		FontIndex = (pCharToWrite - 32) * 6;
		CharacterROMAdress = (unsigned char*)(FontVerdana7 + FontIndex);
		FontWidthIndex = 5;
	}
	else
	{
		FontIndex = (pCharToWrite - 32) * 8;
		CharacterROMAdress = (unsigned char*)(FontVerdana7Bold + FontIndex);
		FontWidthIndex = 7;
	}

	GLCD_WriteData(0x00);

	for(i=0; i<pgm_read_byte(CharacterROMAdress+FontWidthIndex); i++)
	{
		GLCD_WriteData( (pgm_read_byte(CharacterROMAdress + i)) );
	}
}

// Write String with MEDIUM Font size 7
void GLCD_WriteStringM(char* pStringToWrite)
{
	while(*pStringToWrite)
	{
		GLCD_WriteCharM((char)(*pStringToWrite));
		pStringToWrite++;
	}	
}	

// Write String from FLASH with MEDIUM Font size 7
void GLCD_WriteStringM_P(const char* pStringToWrite)
{
	register char Character;

	while( (Character = pgm_read_byte(pStringToWrite)) )
	{
		GLCD_WriteCharM(Character);
		pStringToWrite++;
	}
}	

// Write Integer with MEDIUM Font size 7
void GLCD_WriteNumberM(signed short pNumber)
{
	itoa(pNumber,NumberBuffer,10);
	GLCD_WriteStringM(NumberBuffer);
}


//============================================================================
//				FONT "ZERO TWOS" FOR MITSUBISHI
//============================================================================

void GLCD_WriteDigitZero(unsigned char pDigit,unsigned char pFont)
{
	unsigned short FontIndex;
	unsigned char *CharacterROMAdress;
	unsigned char FontWidthX=0,FontHeightY=0,FontSpace=0,i,Page;
	unsigned char XCursorStart,YCursorStart;

	if(pFont == FONT_ZERO_7)
	{
		FontWidthX = 7;	
		FontHeightY = 1;
		FontSpace = 1;
		CharacterROMAdress = (unsigned char*)(FontZero7);
	}
	else if(pFont == FONT_ZERO_7BOLD)
	{
		FontWidthX = 8;	
		FontHeightY = 1;
		FontSpace = 1;
		CharacterROMAdress = (unsigned char*)(FontZero7Bold);
	}
	else if(pFont == FONT_ZERO_14)
	{
		FontWidthX = 14;	
		FontHeightY = 2;
		FontSpace = 1;
		CharacterROMAdress = (unsigned char*)(FontZero14);
	}
	else if(pFont == FONT_ZERO_22)
	{
		FontWidthX = 22;	
		FontHeightY = 3;
		FontSpace = 2;
		CharacterROMAdress = (unsigned char*)(FontZero22);
	}
	#ifdef CFG_FONT_ZERO30_ENABLE
	else if(pFont == FONT_ZERO_30)
	{
		FontWidthX = 30;	
		FontHeightY = 4;
		FontSpace = 2;
		CharacterROMAdress = (unsigned char*)(FontZero30);
	}
	#endif

	XCursorStart = screen_x;
	YCursorStart = screen_y;
	FontIndex = pDigit * FontWidthX * FontHeightY;
	CharacterROMAdress += FontIndex;

	for(Page=1; Page<(FontHeightY+1); Page++)
	{
		for(i=0; i<FontWidthX; i++) GLCD_WriteData( (pgm_read_byte(CharacterROMAdress++)) );
		if(Page < FontHeightY) GLCD_GoTo(XCursorStart,YCursorStart+Page);
	}

	GLCD_GoTo(screen_x+FontSpace,YCursorStart);
}

void GLCD_WriteNumberZero(unsigned short pNumber,unsigned char pKomma,unsigned char pFont)
{
	unsigned char  NumberIndex = 1;
	unsigned char  NumberArray[7];
	unsigned char  i;
	unsigned char  XBackToKomma = 5;

	     if(pFont == FONT_ZERO_7)		XBackToKomma = 5;
	else if(pFont == FONT_ZERO_7BOLD)	XBackToKomma = 6;
	else if(pFont == FONT_ZERO_14)		XBackToKomma = 10;
	else if(pFont == FONT_ZERO_22)		XBackToKomma = 16;
	#ifdef CFG_FONT_ZERO30_ENABLE
	else if(pFont == FONT_ZERO_30)		XBackToKomma = 22;
	#endif

	// Zahl in Ziffern zerlegen
	do
	{
		NumberArray[NumberIndex] = pNumber%10;
		NumberIndex++;

		pNumber /= 10;
	}
	while(pNumber != 0);
	
	// fhrende Null zwischen 0.1 und 0.9
	if(pKomma == KOMMA && NumberIndex == 2)
	{
		NumberArray[2] = 0;
		NumberIndex = 3;
	}

	// Darstellen
	for(i=NumberIndex-1; i>0; i--)
	{
		if(pKomma == KOMMA && i == 1)
		{
			GLCD_WriteDigitZero(FONT_ZERO_POINT,pFont);
			GLCD_GoTo(screen_x-XBackToKomma,screen_y);
		}
		GLCD_WriteDigitZero(NumberArray[i],pFont);
	}
}


// pXCursor:0..127, pYCursor:0..7
void GLCD_WriteNumberZero_Format(unsigned char pXCursor,unsigned char pYCursor,unsigned char* pNumberAddr,unsigned char pBit,unsigned char pDigits,unsigned char pKomma,unsigned char pFont,unsigned char pDirection)
{	
	unsigned short NewValue=0,OldValue=0;

	if(pBit == _8BIT)
	{
		NewValue = (unsigned char)( *(pNumberAddr) );
		OldValue = (unsigned char)( *(pNumberAddr+1) );
	}
	else if(pBit == _16BIT)
	{
		NewValue = (unsigned short)( *(pNumberAddr)   + 256*(*(pNumberAddr+1)) );
		OldValue = (unsigned short)( *(pNumberAddr+2) + 256*(*(pNumberAddr+3)) );
	}

	if(OldValue != NewValue)
	{
		GLCD_GoTo(pXCursor,pYCursor);

		// Left
		if(pDirection == LEFT_ALIGN) GLCD_WriteNumberZero(NewValue,pKomma,pFont);	
	
		// Overwrite not used number with a space
		if(pDigits >= 5 && NewValue <  10000) GLCD_WriteDigitZero(FONT_ZERO_BLANKO,pFont);
		if(pDigits >= 4 && NewValue <   1000) GLCD_WriteDigitZero(FONT_ZERO_BLANKO,pFont);
		if(pDigits >= 3 && NewValue <    100) GLCD_WriteDigitZero(FONT_ZERO_BLANKO,pFont);
		if(pDigits >= 2 && NewValue <     10 && pKomma == 0) GLCD_WriteDigitZero(FONT_ZERO_BLANKO,pFont);
	
		// Right
		if(pDirection == RIGHT_ALIGN) GLCD_WriteNumberZero(NewValue,pKomma,pFont);	
	
		// Old value = New value
		if(pBit == _8BIT)
		{
			*(pNumberAddr+1) = NewValue;
		}
		else if(pBit == _16BIT)
		{
			*(pNumberAddr+2) = *(pNumberAddr);
			*(pNumberAddr+3) = *(pNumberAddr+1);
		}
	}					
}

// pXCursor:0..127, pYCursor:0..7
void GLCD_WriteNumberZero_SFormat(unsigned char pXCursor,unsigned char pYCursor,signed char* pNumberAddr,unsigned char pBit,unsigned char pDigits,unsigned char pKomma,unsigned char pFont,unsigned char pDirection)
{	
	signed short NewValue=0,OldValue=0,SignMarker=0;

	if(pBit == _8BIT)
	{
		NewValue = (signed char)( *(pNumberAddr) );
		OldValue = (signed char)( *(pNumberAddr+1) );
	}
	else if(pBit == _16BIT)
	{
		NewValue = (signed short)( *(pNumberAddr)   + 256*(*(pNumberAddr+1)) );
		OldValue = (signed short)( *(pNumberAddr+2) + 256*(*(pNumberAddr+3)) );
	}

	if(OldValue != NewValue)
	{
		GLCD_GoTo(pXCursor,pYCursor);

		// Negative Sign
		if(NewValue < 0)
		{
			NewValue = ~NewValue + 1;
			SignMarker = 1;
		}
		
		// Left
		if(pDirection == LEFT_ALIGN)
		{
			if(SignMarker) GLCD_WriteDigitZero(FONT_ZERO_MINUS,pFont);
			GLCD_WriteNumberZero(NewValue,pKomma,pFont);
		}
	
		// Overwrite not used number with a space
		if(pDigits >= 5 && NewValue <  10000) 						GLCD_WriteDigitZero(FONT_ZERO_BLANKO,pFont);
		if(pDigits >= 4 && NewValue <   1000) 						GLCD_WriteDigitZero(FONT_ZERO_BLANKO,pFont);
		if(pDigits >= 3 && NewValue <    100 && !SignMarker)		GLCD_WriteDigitZero(FONT_ZERO_BLANKO,pFont);
		if(pDigits >= 2 && NewValue <     10 && pKomma == NO_KOMMA) GLCD_WriteDigitZero(FONT_ZERO_BLANKO,pFont);
	
		// Right
		if(pDirection == RIGHT_ALIGN)	
		{
			if(SignMarker) GLCD_WriteDigitZero(FONT_ZERO_MINUS,pFont);
			GLCD_WriteNumberZero(NewValue,pKomma,pFont);
		}
	
		// Old value = New value
		if(pBit == _8BIT)
		{
			*(pNumberAddr+1) = NewValue;
		}
		else if(pBit == _16BIT)
		{
			*(pNumberAddr+2) = *(pNumberAddr);
			*(pNumberAddr+3) = *(pNumberAddr+1);
		}
	}					
}


//============================================================================
//				LCD-IMAGES FUNCTIONS
//============================================================================

// pPosX=[0;127] pPosY=[0;7] pHeightX=[0;127] pWidthY=[0;7]
void GLCD_DrawImage(const unsigned char* pImage, unsigned char pPosX, unsigned char pPosY, unsigned char pWidthX, unsigned char pHeightY)
{
	unsigned char PosX=0, PosY=0;	
	unsigned char BlockValue;
	unsigned char BlockSize;
	unsigned short i;


	GLCD_GoTo(pPosX,pPosY);

	do
	{
		if(pgm_read_byte(pImage) == OPTIMIZED_IMAGES_MARKER)
		{
			BlockValue = pgm_read_byte(pImage + 1);
			BlockSize  = pgm_read_byte(pImage + 2);
			
			for(i=0; i<BlockSize; i++)
			{
				GLCD_WriteData(BlockValue);
					
				PosX++;
				if(PosX == pWidthX)
				{
					PosX = 0;
					PosY++;
					GLCD_GoTo(pPosX,pPosY+PosY);
				}	
			}
			pImage += 3;	
		}
		else
		{
			GLCD_WriteData(pgm_read_byte(pImage));
			pImage++;
			
			PosX++;
			if(PosX == pWidthX)
			{
				PosX = 0;
				PosY++;
				GLCD_GoTo(pPosX,pPosY+PosY);
			}
		}		
	}
	while(PosY < pHeightY);
}


// Draw Animation
void GLCD_DrawAni(unsigned short* pAni, unsigned char pFrames, unsigned char pPosX, unsigned char pPosY, unsigned char pWidthX, unsigned char pHeightY, unsigned short pWait)
{
	unsigned char i;
	unsigned char *Frame;


	for(i=0; i<pFrames; i++)
	{
		Frame = (unsigned char*)(*pAni);

		GLCD_DrawImage(Frame,pPosX,pPosY,pWidthX,pHeightY);

		pAni++;

		GLCD_GoTo(pPosX,pPosY);
		WaitInMs(pWait);
	}
}	

// Draw a simple horizontal or vertical line
void GLCD_DrawSimpleLine(unsigned char pPosX1,unsigned char pPosY1,unsigned char pPosX2,unsigned char pPosY2, unsigned char pFillByte)
{
	unsigned char i,Temp;

	if(pPosX1 == pPosX2)							// draw horizontal line at vertical position of LCD-Module
	{
		if(pPosY1 > pPosY2)							// Sort Coords
		{
			Temp = pPosY1;
			pPosY1 = pPosY2;
			pPosY2 = Temp;
		}

		if(pPosY1%8 == 0 && pPosY2%8 == 7)			// it's possible to draw complete 8bit-page of LCD
		{
			for(i=pPosY1/8; i<=pPosY2/8; i++)
			{
				GLCD_GoTo(pPosX1,i);
				GLCD_WriteData(pFillByte);
			}
		}	
		else										// draw a horizontal line consists of pixels instead of 8bit-pages
		{
			for(i=pPosY1; i<=pPosY2; i++)
			{
				GLCD_SetPixel(pPosX1,i,pFillByte);
			}		
		}	
	}	
	else if(pPosY1 == pPosY2)						// draw vertical line at vertical position of LCD-Module
	{
		if(pPosX1 > pPosX2)							// Sort Coords
		{
			Temp = pPosX1;
			pPosX1 = pPosX2;
			pPosX2 = Temp;
		}

		for(i=pPosX1; i<=pPosX2; i++)
		{
			GLCD_SetPixel(i,pPosY1,pFillByte);	
		}		
	}	
}

// Draw a real line
void GLCD_DrawLine(unsigned char x1,unsigned char y1,unsigned char x2,unsigned char y2,unsigned char color,unsigned char pBold)
{
	uint8_t length, i, y, yAlt, xTmp, yTmp;
	int16_t m;
	
	// vertical or horizontal line
	if(x1 == x2 || y1 == y2)
	{
		GLCD_DrawSimpleLine(x1,y1,x2,y2,color);		
	}
	// schiefe line
	else
	{
		// angle > 45
		if(fabs(y2-y1) > fabs(x2-x1))
		{
			// x1 must be smaller than x2
			if(x1 > x2)
			{
				xTmp = x1;
				yTmp = y1;
				x1 = x2;
				y1 = y2;
				x2 = xTmp;
				y2 = yTmp;
			}
		
			length = x2-x1;		// not really the length :)
			m = ((y2-y1)*200)/length;
			yAlt = y1;
			
			for(i=0; i<=length; i++)
			{
				y = ((m*i)/200)+y1;
				
				if((m*i)%200 >= 100)
					y++;
				else if((m*i)%200 <= -100)
					y--;
				
				GLCD_DrawSimpleLine(x1+i, yAlt, x1+i, y, color);

				if(length <= (y2-y1) && y1 < y2)
					yAlt = y+1-pBold;
				else if(length <= (y1-y2) && y1 > y2)
					yAlt = y-1+pBold;
				else
					yAlt = y;
			}
		
		// angle <= 45
		} else
		{
			// y1 must be smaller than y2
			if(y1 > y2)
			{
				xTmp = x1;
				yTmp = y1;
				x1 = x2;
				y1 = y2;
				x2 = xTmp;
				y2 = yTmp;
			}
			
			length = y2-y1;
			m = ((x2-x1)*200)/length;
			yAlt = x1;
			
			for(i=0; i<=length; i++)
			{
				y = ((m*i)/200)+x1;
				
				if((m*i)%200 >= 100)
					y++;
				else if((m*i)%200 <= -100)
					y--;
				
				GLCD_DrawSimpleLine(yAlt, y1+i, y, y1+i, color);

				if(length <= (x2-x1) && x1 < x2)
					yAlt = y+1-pBold;
				else if(length <= (x1-x2) && x1 > x2)
					yAlt = y-1+pBold;
				else
					yAlt = y;
			}
		}
	}
}


//============================================================================
//				SCROLL + SCAN FUNCTIONS
//============================================================================

// Scroll Screen up/down/left/right
#ifdef CFG_SCROLL_SCREEN_ENABLE
void GLCD_ScrollScreen(unsigned char pDirection,unsigned char pCursorX,unsigned char pPageY,unsigned char pWidthX,unsigned char pHeightY,unsigned char pLines,unsigned short pWait)
{
	unsigned char i,x,y,xMax,yMax;
	
	switch(pDirection)
	{
		case SCROLL_RIGHT:
			// Scan display by GLCD_ReadData() and store it to ScreenBuffer[][]
			GLCD_ScanScreen(pCursorX,pPageY,pWidthX,pHeightY);			
									
			xMax = 0;
			for(i=0; i<pLines; i++)
			{
				for(y=0; y<pHeightY; y++)					
				{
					// Print the new Data
					GLCD_GoTo(pCursorX+xMax,y+pPageY);
					GLCD_WriteData(0x00);
					for(x=0; x<pWidthX-xMax-1; x++) GLCD_WriteData(ScreenBuffer[x][y]);
				}
				xMax++;
			
				WaitInMs(pWait);
			}
			break;
			
		case SCROLL_LEFT:
			// Scan display by GLCD_ReadData() and store it to ScreenBuffer[][]
			GLCD_ScanScreen(pCursorX,pPageY,pWidthX,pHeightY);			

			xMax = 0;
			for(i=0; i<pLines; i++)
			{
				for(y=0; y<pHeightY; y++)					
				{
					// Print the new Data
					GLCD_GoTo(pCursorX,y+pPageY);
					for(x=xMax+1; x<pWidthX; x++) GLCD_WriteData(ScreenBuffer[x][y]);
					GLCD_WriteData(0x00);
				}
				xMax++;
			
				WaitInMs(pWait);
			}
			break;

			
		case SCROLL_UP:
			for(i=0; i<pLines; i++)
			{
				// Scan display by GLCD_ReadData() and store it to ScreenBuffer[][]
				GLCD_ScanScreen(pCursorX,pPageY,pWidthX,pHeightY);			
				
				yMax = 0;
				for(y=0; y<pHeightY-yMax-1; y++)					
				{
					// Print the new Data
					GLCD_GoTo(pCursorX,y+pPageY);
					for(x=0; x<pWidthX; x++) GLCD_WriteData(ScreenBuffer[x][y+1]);
				}
				yMax++;

				// Erase the new right line
				GLCD_GoTo(pCursorX,y+pPageY);
				for(x=0; x<pWidthX; x++) GLCD_WriteData(0x00);					
	
				WaitInMs(pWait);
			}	
			break;
			
		case SCROLL_DOWN:
			for(i=0; i<pLines; i++)
			{
				// Scan display by GLCD_ReadData() and store it to ScreenBuffer[][]
				GLCD_ScanScreen(pCursorX,pPageY,pWidthX,pHeightY);			
				
				yMax = 0;

				// Erase the new right line
				GLCD_GoTo(pCursorX,pPageY+yMax);
				for(x=0; x<pWidthX; x++) GLCD_WriteData(0x00);					

				for(y=yMax+1; y<pHeightY; y++)					
				{
					// Print the new Data
					GLCD_GoTo(pCursorX,y+pPageY);
					for(x=0; x<pWidthX; x++) GLCD_WriteData(ScreenBuffer[x][y-1]);
				}
				yMax++;
	
				WaitInMs(pWait);
			}	
			break;
	}	
}	

// Scan the whole Screen; pHalf=1(top screen half); pHalf=0(buttom screen half)
void GLCD_ScanScreen(unsigned char pCursorX,unsigned char pPageY,unsigned char pWidthX,unsigned char pHeightY)
{
	unsigned char x,y;
			
	for(y=0; y<pHeightY; y++)
	{
		for(x=0; x<pWidthX; x++)
		{
			GLCD_GoTo(x+pCursorX,y+pPageY);
			GLCD_ReadData();
			GLCD_GoTo(x+pCursorX,y+pPageY);
			ScreenBuffer[x][y] = GLCD_ReadData();			
		}
	}		
}
#endif
	
